from abc import ABC, abstractmethod
from typing import Optional, Tuple


class Backend(ABC):

    @abstractmethod
    async def process(self, key: str, interval_ms: int, limit: int) -> Tuple[Optional[str], dict]:
        pass


class CountingBackend(Backend):

    async def process(self, key: str, interval_ms: int, limit: int) -> Tuple[Optional[str], dict]:
        requests_made = self.get_and_increment(key, interval_ms, limit)
        meta = {
            'key': key,
            'requests_made': requests_made,
            'max_requests': limit,
            'requests_left': limit - requests_made
        }

        result = None
        if requests_made > limit:
            result = f'Hit limiter {key}: {limit} allowed per {interval_ms}ms'

        return result, meta

    @abstractmethod
    def get_and_increment(self, key: str, interval_ms: int, limit: int) -> int:
        pass



from typing import List

from starlette.requests import Request

from sweelimiter.api.key_extractor import new_extractor


def _host_key_extractor(req: Request) -> List[str]:
    return [req.client.host]


host_key_extractor = new_extractor(
    'host',
    _host_key_extractor
)



from itertools import chain
from typing import Callable, List

from starlette.requests import Request

from .key_extractor import KeyExtractor
from .backend import Backend
from .bypass_checker import BypassChecker
from .limit_config import LimitConfig
from .limiter_result import LimiterResult


class LimiterEngine:

    def __init__(
            self,
            prefix: str,
            extractors: List[KeyExtractor],
            backend: Backend,
            bypass_checkers: List[BypassChecker],
    ):
        self._prefix = prefix
        self._extractors = extractors
        self._backend = backend
        self._bypass_checkers = bypass_checkers

    def _get_all_keys(self, req: Request) -> List[str]:
        keys = list(chain.from_iterable([e.extract(req) for e in self._extractors]))
        return keys

    def _is_bypassed(self, req: Request) -> bool:
        if any(checker(req) for checker in self._bypass_checkers):
            return True
        else:
            return False

    async def limit(self, config: LimitConfig, req: Request) -> LimiterResult:
        if self._is_bypassed(req):
            return LimiterResult(False, 'Bypassed', [])

        limiter_meta = []

        for extractor in self._extractors:
            keys = extractor.extract(req)

            for req_key in keys:

                for limit in config.limits:
                    full_key = f'{self._prefix}@{config.key}@{req_key}@{limit.interval}'

                    error, meta = await self._backend.process(full_key, limit.interval, limit.max_requests)

                    limiter_meta.append(meta)

                    if error is not None:
                        return LimiterResult(True, error, limiter_meta)

        return LimiterResult(False, 'Limiters not hit', limiter_meta)


def create_limiter_engine(
        prefix: str,
        extractors: List[KeyExtractor],
        backend: Backend,
        bypass_checkers: List[BypassChecker] = None,
):
    if bypass_checkers is None:
        bypass_checkers = []

    return LimiterEngine(
        prefix,
        extractors,
        backend,
        bypass_checkers
    )



from setuptools import find_packages, setup

setup(
    name='sweelimiter',
    packages=['sweelimiter', 'sweelimiter.api', 'sweelimiter.backends', 'sweelimiter.extractors'],
    version='0.1.6',
    description='Limiter for fastapi',
    author='Maxim Besogonov',
    author_email='swifstail@gmail.com',
    license='mpl-2.0',
    keywords=['fastapi', 'starlette', 'ratelimit', 'limiter', 'redis'],
    install_requires=['fastapi~=0.68.1', 'starlette~=0.14.2'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest~=6.2.5', 'requests'],
    test_suite='tests',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



from typing import Callable, List

from starlette.requests import Request

KeyExtractorFn = Callable[[Request], List[str]]


class KeyExtractor:
    def __init__(self, prefix: str, fn: KeyExtractorFn):
        self.prefix = prefix
        self.fn = fn

    def extract(self, req: Request) -> List[str]:
        keys = self.fn(req)
        return [f'{self.prefix}@{key}' for key in keys]


def new_extractor(prefix: str, fn: KeyExtractorFn):
    return KeyExtractor(prefix, fn)


from typing import List

from fastapi import Depends, HTTPException
from starlette.requests import Request

from .limit_config import LimitConfig, Limit
from .error_handler import LimiterErrorHandler
from .limiter_engine import LimiterEngine


def create_ratelimit_depends(engine: LimiterEngine, error_handler: LimiterErrorHandler):
    def ratelimit(key: str, *args: Limit) -> Depends:
        config = LimitConfig(key, *args)

        async def call(request: Request):
            limit_result = await engine.limit(config, request)
            print(limit_result)

            if limit_result.limit_applied:
                if error_handler:
                    error_handler(limit_result)
                else:
                    raise HTTPException(status_code=429, detail=limit_result.description)

        return Depends(call)

    return ratelimit



from typing import Optional, Tuple

from sweelimiter.api.backend import Backend

try:
    from redis import Redis
except ImportError:
    raise AssertionError('Please install redis to use redis_backend')


class RedisBackend(Backend):

    async def process(self, key: str, interval_ms: int, limit: int) -> Tuple[Optional[str], dict]:
        redis_key = 'limiter_' + key
        r = self.r

        seconds, microseconds = r.time()
        time_now = int(seconds * 1000 + microseconds / 1000)

        separation = round(interval_ms / limit)

        r.setnx(redis_key, 0)
        tat = max(int(r.get(redis_key)), time_now)

        wait_left = tat - time_now
        est_interval = interval_ms - separation

        if wait_left <= est_interval:
            new_tat = max(tat, time_now) + separation
            r.set(redis_key, new_tat)
            return (None, {
                'key': key,
                'tat': new_tat,
                'separation': separation,
                'wait': wait_left,
                'est_interval': est_interval
            })

        return (f'Hit limiter {key}: {limit} allowed per {interval_ms}ms, ', {
            'key': key,
            'tat': tat,
            'separation': separation,
            'wait': wait_left,
            'est_interval': est_interval
        })

    def __init__(self, r: Redis):
        self.r = r



from functools import wraps
from typing import List

from fastapi import HTTPException
from starlette.requests import Request

from .error_handler import LimiterErrorHandler
from .limiter_engine import LimiterEngine
from .limit_config import LimitConfig, Limit


def create_ratelimit_decorator(engine: LimiterEngine, error_handler: LimiterErrorHandler):
    def ratelimit(key: str, limits: List[Limit]):
        config = LimitConfig(key, limits)

        def decorator(func):
            @wraps(func)
            async def wrapper(request: Request, *args, **kwargs):
                limit_result = await engine.limit(config, request)
                print(limit_result)

                if limit_result.limit_applied:
                    if error_handler:
                        error_handler(limit_result)
                    else:
                        raise HTTPException(status_code=429, detail=limit_result.description)

                await func(request, *args, **kwargs)

            return wrapper

        return decorator

    return ratelimit
